# -*- coding: utf-8 -*-

from torch import isnan,tensor
from torch import nn, Tensor
from abc import abstractmethod
import logging

logger = logging.getLogger(__name__)


class BaseTNorm(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def __call__(self,
              a: Tensor,
              b: Tensor) -> Tensor:
        raise NotImplementedError

    def conorm(self,
               a: Tensor,
               b: Tensor) -> Tensor:
        return 1.0 - self(1.0 - a, 1.0 - b)

    def adjust_nan_param(self):
        for key, value in self.__dict__.items():
            if type(value) == nn.Parameter:
                if isnan(value):
                    print("The Value of the paramter became NAN")
                    value.data = tensor(0.1, device = value.device)



# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor

from torchnorms.tnorms.base import BaseTNorm

from typing import Optional


class AczelAlsinaTNorm(BaseTNorm):
    def __init__(self,
                 p: Optional[Tensor] = None,
                 default_p: float = 0.1) -> None:
        super().__init__()
        self.p = p
        if self.p is None:
            self.p = nn.Parameter(torch.tensor(default_p))

        else:
            assert self.p > 0
            assert self.p != float('inf')

        assert len(self.p.shape) == 0

        self.relu = torch.nn.ReLU()
        self.eps = 0.001

    def __call__(self,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        self.p.data = self.relu(self.p.data)
        if self.p == 0:
            self.p.data += self.eps

        #Taking the log here might be problamitc as some scores might be explicit 0's
        # Maybe we need to ofset by a small eps. i.e. 10^-4 ?
        p_1 = torch.pow(torch.abs(torch.log(a)), self.p)
        p_2 = torch.pow(torch.abs(torch.log(b)), self.p)
        res = torch.exp(-torch.pow(p_1 + p_2, 1.0 / self.p))
        return res



# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor

from torchnorms.tconorms.base import BaseTCoNorm

from typing import Optional


class DuboisTCoNorm(BaseTCoNorm):
	def __init__(self,
				 p: Optional[Tensor] = None,
				 default_p: float = 0.1) -> None:
		super().__init__()
		self.p = p
		if self.p is None:
			assert default_p >= -1.0

			self.p = nn.Parameter(torch.tensor(default_p))
		else:
			assert self.p >= -1.0

		assert len(self.p.shape) == 0

	def __call__(self,
				 a: Tensor,
				 b: Tensor) -> Tensor:

			self.adjust_nan_param()

			self.p.data = torch.clamp(self.p.data, min=0, max=1)

			p_1 = a + b
			p_2 = p_1 + (self.p * a * b)
			res = torch.minimum(p_2, torch.tensor(1))

			return res



# -*- coding: utf-8 -*-

import torch
from typing import Optional
from torch import nn, Tensor
from torchnorms.negations.base import BaseNegation


class YagerNegation(BaseNegation):
	def __init__(self,
				 p: Optional[Tensor] = None,
				 default_p: float = 1.1) -> None:
		super().__init__()
		self.p = p
		if self.p is None:
			assert default_p > -1
			self.p = nn.Parameter(torch.tensor(default_p))
		assert len(self.p.shape) == 0

		self.relu = torch.nn.ReLU()

		self.eps = 0.01
		self.__name__ = 'Yager'

	def __call__(self, a: Tensor) -> Tensor:

		self.adjust_nan_param()

		self.p.data = self.relu(self.p.data)
		if self.p == 0:
			self.p.data += self.eps

		res = torch.pow((1.0 - torch.pow(a, self.p)), 1.0 / self.p)

		return res



# -*- coding: utf-8 -*-

from torchnorms.tnorms.base import BaseTNorm
from torchnorms.tnorms.aa import AczelAlsinaTNorm
from torchnorms.tnorms.additive_generator import AdditiveLinearGenerator, AdditiveGenerator, AdditiveOrthogonalGenerator, AdditiveLearnableInverseGenerator, AdditivePolynomialGenerator
from torchnorms.tnorms.classic import MinimumTNorm, ProductTNorm, LukasiewiczTNorm, DrasticTNorm, EinsteinTNorm, BoundedTNorm, NilpotentMinimumTNorm, HamacherProductTNorm
from torchnorms.tnorms.dombi import DombiTNorm
from torchnorms.tnorms.dubois import DuboisTNorm
from torchnorms.tnorms.hamacher import HamacherTNorm
from torchnorms.tnorms.ss import SchweizerSklarTNorm
from torchnorms.tnorms.weber import WeberTNorm
from torchnorms.tnorms.yager import YagerTNorm
from torchnorms.tnorms.frank import FrankTNorm

__all__ = [
    'BaseTNorm',
    'AczelAlsinaTNorm',
    'AdditiveLinearGenerator',
    'AdditiveGenerator',
    'AdditiveOrthogonalGenerator',
    'AdditiveLearnableInverseGenerator',
    'AdditivePolynomialGenerator',
    'MinimumTNorm',
    'ProductTNorm',
    'LukasiewiczTNorm',
    'DrasticTNorm',
    'EinsteinTNorm',
    'BoundedTNorm',
    'NilpotentMinimumTNorm',
    'HamacherTNorm',
    'DombiTNorm',
    'DuboisTNorm',
    'HamacherProductTNorm',
    'SchweizerSklarTNorm',
    'WeberTNorm',
    'YagerTNorm',
    'FrankTNorm'
]



# -*- coding: utf-8 -*-

import torch
from torch import Tensor
from torchnorms.negations.base import BaseNegation


class StandardNegation(BaseNegation):
    def __init__(self) -> None:
        super().__init__()
        self.__name__ = 'standard'

    @classmethod
    def __call__(cls,
                 a: Tensor) -> Tensor:
        return 1.0 - a


class StrictNegation(BaseNegation):
    def __init__(self) -> None:
        super().__init__()
        self.__name__ = 'strict'

    @classmethod
    def __call__(cls,
                 a: Tensor) -> Tensor:
        return 1.0 - torch.pow(a, 2)


class StrictCosNegation(BaseNegation):
    def __init__(self) -> None:
        super().__init__()
        self.__name__ = 'strict_cosine'

    @classmethod
    def __call__(cls,
                 a: Tensor) -> Tensor:
        pi = torch.acos(torch.zeros(1)).item() * 2
        res = 0.5 * (torch.cos(pi * a))
        return res



# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor

from torchnorms.tnorms.base import BaseTNorm

from typing import Optional


class DombiTNorm(BaseTNorm):
	def __init__(self,
				 p: Optional[Tensor] = None,
				 default_p: float = 0.1) -> None:
		super().__init__()
		self.p = p
		if self.p is None:
			self.p = nn.Parameter(torch.tensor(default_p))
		assert len(self.p.shape) == 0
		self.relu = torch.nn.ReLU()
		self.eps = 0.001

	def __call__(self,
				 a: Tensor,
				 b: Tensor) -> Tensor:

		self.adjust_nan_param()

		self.p.data = self.relu(self.p.data)

		if self.p == 0:
			# eps is added to avoid 0
			self.p.data += self.eps

		p_1 = torch.pow((1 - a) / a, self.p)
		p_2 = torch.pow((1 - b) / b, self.p)
		p_3 = 1.0 + torch.pow(p_1 + p_2, 1.0 / self.p)
		res = 1.0 / p_3

		return res



# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor

from torchnorms.tconorms.base import BaseTCoNorm

from typing import Optional


class YagerTCoNorm(BaseTCoNorm):
    def __init__(self,
                 p: Optional[Tensor] = None,
                 default_p: float = 0.1) -> None:
        super().__init__()
        self.p = p
        if self.p is None:
            self.p = nn.Parameter(torch.tensor(default_p))

        assert len(self.p.shape) == 0

        self.relu = torch.nn.ReLU()
        self.eps = 0.001

    def __call__(self,
                 a: Tensor,
                 b: Tensor) -> Tensor:

            res: Optional[Tensor] = None

            self.adjust_nan_param()

            self.p = self.relu(self.p)
            if self.p == 0:
                self.p += self.eps

            p_1 = torch.pow(a, self.p)
            p_2 = torch.pow(b, self.p)
            p_3 = torch.pow(p_1 + p_2, 1.0 / self.p)
            res = torch.minimum(torch.tensor(1.0) , p_3)

            return res



# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor

from torchnorms.tconorms.base import BaseTCoNorm

from typing import Optional


class SchweizerSklarTCoNorm(BaseTCoNorm):
    def __init__(self,
                 p: Optional[Tensor] = None,
                 default_p: float = 0.1) -> None:
        super().__init__()
        self.p = p
        if self.p is None:
            self.p = nn.Parameter(torch.tensor(default_p))

        assert len(self.p.shape) == 0

        self.eps += 0.001

    def __call__(self,
                 a: Tensor,
                 b: Tensor) -> Tensor:

            self.adjust_nan_param()

            if self.p == 0:
                self.p.data += self.eps

            p_1 = torch.pow(1.0 - a, self.p)
            p_2 = torch.pow(1.0 - b, self.p)
            p_1 = torch.maximum(torch.tensor(0), p_1 + p_2 - 1.0)
            res = torch.pow(p_1, 1.0 / self.p)
            res = 1.0 - res

            return res



# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor

from torchnorms.tnorms.base import BaseTNorm


class NeuralSiameseTNorm(BaseTNorm):
    def __init__(self) -> None:
        super().__init__()
        self.model = nn.Linear(in_features=1, out_features=1, bias=True)

    def __call__(self,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        return self.model(a) * self.model(b)


class NeuralTNorm(BaseTNorm):
    def __init__(self) -> None:
        super().__init__()
        self.model = nn.Linear(in_features=2, out_features=1, bias=True)

    def __call__(self,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        return self.model(a) * self.model(b)



# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor

from torchnorms.tnorms.base import BaseTNorm

from typing import Optional


class DuboisTNorm(BaseTNorm):
    def __init__(self,
                 p: Optional[Tensor] = None,
                 default_p: float = 0.1) -> None:
        super().__init__()
        self.p = p
        if self.p is None:
            self.p = nn.Parameter(torch.tensor(default_p))
        assert len(self.p.shape) == 0

    def __call__(self,
                 a: Tensor,
                 b: Tensor) -> Tensor:

        self.adjust_nan_param()

        self.p.data = torch.clamp(self.p.data, min=0, max=1)
        res = (a * b) / torch.maximum(torch.maximum(a, b), self.p)

        return res
